from setuptools import setup
import setuptools

setup(name='gymutils',
      version='0.0.1',
      url='https://github.com/dennybritz/reinforcement-learning',
      author='Ryan McDermott (forked from Denny Britz)',
      packages=setuptools.find_packages(),
      install_requires=['numpy']
      )
